""" Vault interactions """
from __future__ import print_function
import os
import socket
import hvac
import yaml
# need to override those SSL warnings
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from aomi.helpers import problems, log, cli_hash, merge_dicts
import aomi.seed
from aomi.template import render, load_var_files

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


def app_token(vault_client, app_id, user_id):
    """Returns a vault token based on the app and user id."""
    resp = vault_client.auth_app_id(app_id, user_id)
    if 'auth' in resp and 'client_token' in resp['auth']:
        return resp['auth']['client_token']
    else:
        problems('Unable to retrieve app token')


def initial_token(vault_client, opt):
    """Generate our first token based on workstation configuration"""
    token_file = os.environ.get('VAULT_TOKEN_FILE',
                                "%s/.vault-token" % os.environ['HOME'])
    app_file = os.environ.get('AOMI_APP_FILE',
                              "%s/.aomi-app-token" % os.environ['HOME'])
    token_file = os.path.abspath(token_file)
    app_file = os.path.abspath(app_file)
    if 'VAULT_TOKEN' in os.environ and len(os.environ['VAULT_TOKEN']) > 0:
        log('Token derived from VAULT_TOKEN environment variable', opt)
        return os.environ['VAULT_TOKEN'].strip()
    elif 'VAULT_USER_ID' in os.environ and \
         'VAULT_APP_ID' in os.environ and \
         len(os.environ['VAULT_USER_ID']) > 0 and \
         len(os.environ['VAULT_APP_ID']) > 0:
        token = app_token(vault_client,
                          os.environ['VAULT_APP_ID'].strip(),
                          os.environ['VAULT_USER_ID'].strip())
        log("Token derived from VAULT_APP_ID and VAULT_USER_ID", opt)
        return token
    elif os.path.exists(app_file):
        token = yaml.load(open(app_file).read().strip())
        if 'app_id' in token and 'user_id' in token:
            token = app_token(vault_client,
                              token['app_id'],
                              token['user_id'])
            log("Token derived from %s" % app_file, opt)
            return token
    elif os.path.exists(token_file):
        log("Token derived from %s" % token_file, opt)
        return open(token_file, 'r').read().strip()
    else:
        problems('Unable to determine vault authentication method')


def token_meta(operation, opt):
    """Generates metadata for a token"""
    meta = {
        'operation': operation,
        'hostname': socket.gethostname()
    }
    if 'USER' in os.environ:
        meta['unix_user'] = os.environ['USER']

    if opt.metadata:
        meta_bits = opt.metadata.split(',')
        for meta_bit in meta_bits:
            k, v = meta_bit.split('=')

        if k not in meta:
            meta[k] = v

    for k, v in meta.items():
        log("Token metadata %s %s" % (k, v), opt)

    return meta


def operational_token(vault_client, operation, opt):
    """Return a properly annotated token for our use."""
    args = {
        'lease': opt.lease,
        'display_name': 'aomi token',
        'meta': token_meta(operation, opt)
    }
    token = vault_client.create_token(**args)
    log("Using lease of %s" % opt.lease, opt)
    return token['auth']['client_token']


def client(operation, opt):
    """Return a vault client"""
    if 'VAULT_ADDR' not in os.environ:
        problems('VAULT_ADDR must be defined')

    vault_host = os.environ['VAULT_ADDR']

    ssl_verify = True
    if 'VAULT_SKIP_VERIFY' in os.environ:
        if os.environ['VAULT_SKIP_VERIFY'] == '1':
            log('Skipping SSL Validation!', opt)
            ssl_verify = False

    log("Connecting to %s" % vault_host, opt)
    vault_client = hvac.Client(vault_host, verify=ssl_verify)
    vault_client.token = initial_token(vault_client, opt)
    if not vault_client.is_authenticated():
        problems("Unable to retrieve initial token")

    vault_client.token = operational_token(vault_client, operation, opt)
    if not vault_client.is_authenticated():
        problems("Unable to retrieve operational token")

    return vault_client


def get_secretfile(opt):
    """Renders, YAMLs, and returns the Secretfile construct"""
    secretfile_path = os.path.abspath(opt.secretfile)
    obj = merge_dicts(load_var_files(opt),
                      cli_hash(opt.extra_vars))
    return yaml.load(render(secretfile_path, obj))


def seed_secrets(config, vault_client, opt):
    """Seed our various secrets"""
    for secret in config.get('secrets', []):
        if 'var_file' in secret:
            aomi.seed.var_file(vault_client, secret, opt)
        elif 'aws_file' in secret:
            aomi.seed.aws(vault_client, secret, opt)
        elif 'files' in secret:
            aomi.seed.files(vault_client, secret, opt)
        else:
            problems("Invalid secret element %s" % secret, vault_client)


def seed(vault_client, opt):
    """Will provision vault based on the definition within a Secretfile"""
    config = get_secretfile(opt)
    seed_secrets(config, vault_client, opt)

    for policy in config.get('policies', []):
        if 'name' in policy:
            aomi.seed.policy(vault_client, policy, opt)
        else:
            problems('Invalid policy %s' % policy, vault_client)

    for app in config.get('apps', []):
        if 'app_file' in app:
            aomi.seed.app(vault_client, app, opt)
        else:
            problems("Invalid app element %s" % app, vault_client)

    for users in config.get('users', []):
        aomi.seed.users(vault_client, users, opt)

    for audit_log in config.get('audit_logs', []):
        aomi.seed.audit_logs(vault_client, audit_log, opt)



""" Secret rendering """
import os
from pkg_resources import resource_filename
from aomi.helpers import problems, warning, cli_hash, merge_dicts
from aomi.template import render, load_var_files


def secret_key_name(path, key, opt):
    """Renders a Secret key name appropriately"""
    value = key
    if opt.merge_path:
        norm_path = [x for x in path.split('/') if x]
        value = "%s_%s" % ('_'.join(norm_path), key)

    if opt.add_prefix:
        value = "%s%s" % (opt.add_prefix, value)

    if opt.add_suffix:
        value = "%s%s" % (value, opt.add_suffix)

    return value


def grok_template_file(src):
    """Determine the real deal template file"""
    if not src.startswith('builtin:'):
        return os.path.abspath(src)
    else:
        builtin = src.split(':')[1]
        builtin = "templates/%s.j2" % builtin
        return resource_filename(__name__, builtin)


def blend_vars(secrets, opt):
    """Blends secret and static variables together"""
    extra_obj = merge_dicts(load_var_files(opt),
                            cli_hash(opt.extra_vars))
    merged = merge_dicts(extra_obj, secrets)
    template_obj = dict((k, v) for k, v in merged.iteritems() if v)
    # give templates something to iterate over
    template_obj['aomi_items'] = template_obj.copy()
    return template_obj


def template(client, src, dest, paths, opt):
    """Writes a template using variables from a vault path"""
    key_map = cli_hash(opt.key_map)
    obj = {}
    for path in paths:
        data = client.read(path)['data']
        for s_k, s_v in data.items():
            o_key = s_k
            if s_k in key_map:
                o_key = key_map[s_k]

            k_name = secret_key_name(path, o_key, opt) \
                .lower() \
                .replace('-', '_')
            obj[k_name] = s_v

    template_obj = blend_vars(obj, opt)
    output = render(grok_template_file(src),
                    template_obj)
    open(os.path.abspath(dest), 'w').write(output)


def raw_file(client, src, dest):
    """Write the contents of a vault path/key to a file"""
    path_bits = src.split('/')
    path = '/'.join(path_bits[0:len(path_bits) - 1])
    key = path_bits[len(path_bits) - 1]
    resp = client.read(path)
    if not resp:
        problems("Unable to retrieve %s" % path, client)
    else:
        if 'data' in resp and key in resp['data']:
            secret = resp['data'][key]
            open(os.path.abspath(dest), 'w').write(secret)
        else:
            problems("Key %s not found in %s" % (key, path), client)


def env(client, paths, opt):
    """Renders a shell snippet based on paths in a Secretfile"""
    old_prefix = False
    old_prefix = opt.prefix and not (opt.add_prefix or
                                     opt.add_suffix or
                                     not opt.merge_path)
    if old_prefix:
        warning("the prefix option is deprecated but being used "
                "due to not passing in new options")
    elif opt.prefix:
        warning("the prefix option is deprecated but not being "
                "used due to passing in new options")
    key_map = cli_hash(opt.key_map)
    for path in paths:
        secrets = client.read(path)
        if secrets and 'data' in secrets:
            for s_key, s_val in secrets['data'].items():
                o_key = s_key
                if s_key in key_map:
                    o_key = key_map[s_key]

                # see https://github.com/Autodesk/aomi/issues/40
                env_name = None
                if old_prefix:
                    env_name = ("%s_%s" % (opt.prefix, o_key)).upper()
                else:
                    env_name = secret_key_name(path, o_key, opt).upper()

                print("%s=\"%s\"" % (env_name, s_val))
                if opt.export:
                    print("export %s" % env_name)


def grok_seconds(lease):
    """Ensures that we are returning just seconds"""
    if lease.endswith('s'):
        return int(lease[0:-1])
    elif lease.endswith('m'):
        return int(lease[0:-1] * 60)
    elif lease.endswith('h'):
        return int(lease[0:-1] * 3600)
    else:
        return None


def aws(client, path, opt):
    """Renders a shell environment snippet with AWS information"""
    creds = client.read(path)
    seconds = grok_seconds(opt.lease)
    if not seconds:
        problems("Passed in invalid lease type %s" % opt.lease, client)

    renew = client.renew_secret(creds['lease_id'], seconds)
    # sometimes it takes a bit for vault to respond
    # if we are within 5s then we are fine
    if seconds - renew['lease_duration'] >= 5:
        problems('Unable to renew secret with desired lease', client)

    if creds and 'data' in creds:
        print("AWS_ACCESS_KEY_ID=\"%s\"" % creds['data']['access_key'])
        print("AWS_SECRET_ACCESS_KEY=\"%s\"" % creds['data']['secret_key'])
        if 'security_token' in creds['data'] \
           and creds['data']['security_token']:
            token = creds['data']['security_token']
            print("AWS_SECURITY_TOKEN=\"%s\"" % token)
    else:
        problems("Unable to generate AWS credentials from %s" % path, client)

    if opt.export:
        print("export AWS_ACCESS_KEY_ID")
        print("export AWS_SECRET_ACCESS_KEY")
        if 'security_token' in creds['data'] \
           and creds['data']['security_token']:
            print("export AWS_SECURITY_TOKEN")



import os
import re
import yaml
import hvac
from aomi.helpers import problems, hard_path, log, is_tagged, warning
import aomi.validation


def sanitize_mount(mount):
    """Returns a quote-unquote sanitized mount path"""
    sanitized_mount = mount
    if sanitized_mount.startswith('/'):
        sanitized_mount = sanitized_mount[1:]

    if sanitized_mount.endswith('/'):
        sanitized_mount = sanitized_mount[:-1]

    return sanitized_mount


def is_mounted(mount, backends, style):
    """Determine whether a backend of a certain type is mounted"""
    for m, v in backends.items():
        b_norm = '/'.join([x for x in m.split('/') if x])
        m_norm = '/'.join([x for x in mount.split('/') if x])
        if (m_norm == b_norm) and v['type'] == style:
            return True

    return False


def ensure_mounted(client, backend, mount):
    """Will ensure a mountpoint exists, or bail with a polite error"""
    backends = client.list_secret_backends()
    if not is_mounted(mount, backends, backend):
        try:
            client.enable_secret_backend(backend, mount_point=mount)
        except hvac.exceptions.InvalidRequest as e:
            m = re.match('existing mount at (?P<path>.+)', str(e))
            if m:
                problems("%s has a mountpoint conflict with %s" %
                         (mount, m.group('path')))
            else:
                raise e


def ensure_auth(client, auth):
    """Will ensure a particular auth endpoint is mounted"""
    backends = client.list_auth_backends().keys()
    backends = [x.rstrip('/') for x in backends]
    if auth not in backends:
        client.enable_auth_backend(auth)


def var_file(client, secret, opt):
    """Seed a var_file into Vault"""
    aomi.validation.var_file_obj(secret)
    my_mount = sanitize_mount(secret['mount'])
    path = "%s/%s" % (my_mount, secret['path'])
    var_file_name = hard_path(secret['var_file'], opt.secrets)
    aomi.validation.secret_file(var_file_name)
    varz = yaml.load(open(var_file_name).read())

    if not aomi.validation.tag_check(secret, path, opt):
        return

    ensure_mounted(client, 'generic', my_mount)

    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    client.write(path, **varz)
    log('wrote var_file %s into %s/%s' % (
        var_file_name,
        my_mount,
        secret['path']), opt)


def aws_region(secret, aws_obj):
    """Return the AWS region with appropriate output"""
    if 'region' in secret:
        return secret['region']
    elif 'region' in aws_obj:
        # see https://github.com/Autodesk/aomi/issues/40
        warning('Defining region in the AWS yaml is deprecated')
        return aws_obj['region']
    else:
        problems('AWS region is not defined')


def aws_roles(secret, aws_obj):
    """Return the AWS roles with appropriate output"""
    if 'roles' in secret:
        return secret['roles']
    elif 'roles' in aws_obj:
        # see https://github.com/Autodesk/aomi/issues/40
        warning('Defining roles within the AWS yaml is deprecated')
        return aws_obj['roles']
    else:
        problems('No AWS roles defined')


def aws(client, secret, opt):
    """Seed an aws_file into Vault"""
    aomi.validation.aws_file_obj(secret)

    my_mount = sanitize_mount(secret['mount'])
    aws_path = "%s/config/root" % my_mount
    if not aomi.validation.tag_check(secret, aws_path, opt):
        return

    ensure_mounted(client, 'aws', my_mount)

    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    aws_file_path = hard_path(secret['aws_file'], opt.secrets)
    aomi.validation.secret_file(aws_file_path)

    aws_obj = yaml.load(open(aws_file_path, 'r').read())
    aomi.validation.aws_secret_obj(aws_file_path, aws_obj)

    region = aws_region(secret, aws_obj)

    obj = {
        'access_key': aws_obj['access_key_id'],
        'secret_key': aws_obj['secret_access_key'],
        'region': region
    }
    client.write(aws_path, **obj)
    log('wrote aws secrets %s into %s' % (
        aws_file_path,
        aws_path), opt)

    ttl_obj = {}
    lease_msg = ''
    if 'lease' in secret:
        ttl_obj['lease'] = secret['lease']
        lease_msg = "lease:%s" % (ttl_obj['lease'])

    if 'lease_max' in secret:
        ttl_obj['lease_max'] = secret['lease_max']
    else:
        if 'lease' in ttl_obj:
            ttl_obj['lease_max'] = ttl_obj['lease']

    if lease_msg == '':
        if 'lease' in aws_obj:
            ttl_obj['lease'] = aws_obj['lease']
            lease_msg = "lease:%s" % (ttl_obj['lease'])

        if 'lease_max' in aws_obj:
            ttl_obj['lease_max'] = aws_obj['lease_max']
        else:
            if 'lease' in ttl_obj:
                ttl_obj['lease_max'] = ttl_obj['lease']

        if lease_msg != '':
            # see https://github.com/Autodesk/aomi/issues/40
            warning('Setting lease and lease_max from the '
                    'AWS yaml is deprecated')

    if 'lease_max' in ttl_obj:
        lease_msg = "%s lease_max:%s" % (lease_msg, ttl_obj['lease_max'])

    if ttl_obj:
        client.write("%s/config/lease" % (secret['mount']), **ttl_obj)
        log("Updated lease for %s %s" % (secret['mount'], lease_msg), opt)

    roles = aws_roles(secret, aws_obj)

    seed_aws_roles(client, secret['mount'], roles, opt)


def seed_aws_roles(client, mount, roles, opt):
    for role in roles:
        if 'name' not in role or \
           ('policy' not in role and 'arn' not in role):
            problems("Invalid role definition %s" % role)

        role_path = "%s/roles/%s" % (mount, role['name'])
        if 'policy' in role:
            data = open(hard_path(role['policy'], opt.policies), 'r').read()
            client.write(role_path, policy=data)
        elif 'arn' in role:
            client.write(role_path, arn=role['arn'])


def app_users(client, app_id, p_users):
    """Write out users for an application"""
    for user in p_users:
        if 'id' not in user:
            problems("Invalid user definition %s" % user)

        user_path = "auth/app-id/map/user-id/%s" % user['id']
        user_obj = {'value': app_id}
        if 'cidr' in user:
            user_obj['cidr_block'] = user['cidr']

        client.write(user_path, **user_obj)


def app(client, app_obj, opt):
    """Seed an app file into Vault"""
    if 'app_file' not in app_obj:
        problems("Invalid app definition %s" % app_obj)

    name = None
    if 'name' in app_obj:
        name = app_obj['name']
    else:
        name = os.path.splitext(os.path.basename(app_obj['app_file']))[0]

    if not aomi.validation.tag_check(app_obj, "app-id/%s" % name, opt):
        return

    app_file = hard_path(app_obj['app_file'], opt.secrets)
    aomi.validation.secret_file(app_file)
    data = yaml.load(open(app_file).read())

    ensure_auth(client, 'app-id')
    if opt.mount_only:
        log("Only enabling app-id", opt)
        return

    if 'users' not in data:
        problems("Invalid app file %s" % app_file)

    policy_name = None
    if 'policy_name' in data:
        warning('Defining policy_name within the app yaml is deprecated')
        policy_name = data['policy_name']
    elif 'policy_name' in app_obj:
        policy_name = app_obj['policy_name']
    else:
        policy_name = name

    app_id = None
    if 'app_id' in data:
        warning('Defining app_id within the app yaml is deprecated')
        app_id = data['app_id']
    elif 'app_id' in app_obj:
        app_id = app_obj['app_id']
    else:
        app_id = name

    policy_file = None
    if 'policy' in data:
        warning('Defining policy_name within the app yaml is deprecated')
        policy_file = data['policy']
    elif 'policy' in app_obj:
        policy_file = app_obj['policy']

    if policy_file:
        policy_data = open(hard_path(policy_file, opt.policies), 'r').read()
        if policy_name in client.list_policies():
            if policy_data != client.get_policy(policy_name):
                problems("Policy %s already exists and content differs"
                         % policy_name)

        log("Using inline policy %s" % policy_name, opt)
        client.set_policy(name, policy_data)
    else:
        if policy_name not in client.list_policies():
            problems("Policy %s is not inline but does not exist"
                     % policy_name)

        log("Using existing policy %s" % policy_name, opt)

    app_path = "auth/app-id/map/app-id/%s" % app_id
    app_obj = {'value': policy_name, 'display_name': name}
    client.write(app_path, **app_obj)
    r_users = data.get('users', [])
    app_users(client, app_id, r_users)
    log('created %d users in application %s' % (len(r_users), name), opt)


def files(client, secret, opt):
    """Seed files into Vault"""
    aomi.validation.file_obj(secret)
    obj = {}
    my_mount = sanitize_mount(secret['mount'])
    vault_path = "%s/%s" % (my_mount, secret['path'])
    if not aomi.validation.tag_check(secret, vault_path, opt):
        return
    ensure_mounted(client, 'generic', my_mount)
    if opt.mount_only:
        log("Only mounting %s" % my_mount, opt)
        return

    for f in secret.get('files', []):
        if 'source' not in f or 'name' not in f:
            problems("Invalid file specification %s" % f)

        filename = hard_path(f['source'], opt.secrets)
        aomi.validation.secret_file(filename)
        data = open(filename, 'r').read()
        obj[f['name']] = data
        log('writing file %s into %s/%s' % (
            filename,
            vault_path,
            f['name']), opt)

    client.write(vault_path, **obj)


def policy(client, secret, opt):
    """Seed a standalone policy into Vault"""
    aomi.validation.policy_obj(secret)

    policy_name = secret['name']
    if not aomi.validation.tag_check(secret, "app-id/%s" % policy_name, opt):
        return

    policy_data = open(hard_path(secret['file'], opt.policies), 'r').read()
    log('writing policy %s' % policy_name, opt)
    client.set_policy(policy_name, policy_data)


def audit_logs(client, log_obj, opt):
    """Creates an audit log entry in Vault"""
    aomi.validation.audit_log_obj(log_obj)
    log_type = log_obj['type']
    vault_path = log_obj.get('path', log_type)
    if not aomi.validation.tag_check(log_obj, vault_path, opt):
        return

    existing_backends = client.list_audit_backends()['data']
    audit_path = "%s/" % vault_path
    if audit_path in existing_backends.keys():
        if log_type == existing_backends[audit_path]['type']:
            log("audit log %s already exists" % vault_path, opt)
            return
        else:
            problems("conflicting audit log at %s" % vault_path)

    obj = {
        'type': log_obj['type']
    }
    if log_type == 'file':
        obj['file_path'] = log_obj['file_path']

    if log_type == 'syslog':
        if 'tag' in log_obj:
            obj['tag'] = log_obj['tag']

        if 'facility' in log_obj:
            obj['facility'] = log_obj['facility']

    if 'description' in log_obj:
        client.enable_audit_backend(log_type,
                                    description=log_obj['description'],
                                    options=obj,
                                    name=vault_path)
    else:
        client.enable_audit_backend(log_type,
                                    options=obj,
                                    name=vault_path)

    log("created %s audit log at %s" % (log_type, vault_path), opt)


def users(client, user_obj, opt):
    """Creates userpass users in Vault"""
    aomi.validation.user_obj(user_obj)

    name = user_obj['username']

    if not aomi.validation.tag_check(user_obj,
                                     "userpass/%s" % name,
                                     opt):
        return

    password_file = hard_path(user_obj['password_file'],
                              opt.secrets)
    aomi.validation.secret_file(password_file)
    password = open(password_file).readline().strip()

    ensure_auth(client, 'userpass')

    user_path = "auth/userpass/users/%s" % name
    v_obj = {
        'password': password,
        'policies': ','.join(user_obj['policies'])
    }
    client.write(user_path, **v_obj)



""" Helpers for aomi that are used throughout the application """
from __future__ import print_function
import sys
import os
from pkg_resources import resource_string, resource_filename


def my_version():
    """Return the version, checking both packaged and development locations"""
    if os.path.exists(resource_filename(__name__, 'version')):
        return resource_string(__name__, 'version')
    else:
        return open("%s/../version" % os.path.dirname(__file__)).read()

VERSION = my_version()


def log(msg, opt):
    """Verbose messaging!"""
    if opt.verbose:
        print(msg, file=sys.stderr)


def warning(msg):
    """Print a warning message to stderr"""
    print("Warning {0}".format(msg), file=sys.stderr)


def problems(msg, client=None):
    """Simple give-up and error out function."""
    if client:
        client.revoke_self_token()

    print("Problem: %s" % msg,
          file=sys.stderr)
    exit(1)


def abspath(raw):
    """Return what is hopefully a OS independent path."""
    path_bits = []
    if raw.find('/') != -1:
        path_bits = raw.split('/')
    elif raw.find('\\') != -1:
        path_bits = raw.split('\\')
    else:
        path_bits = [raw]

    return os.path.abspath(os.sep.join(path_bits))


def hard_path(path, prefix_dir):
    """Returns an absolute path to either the relative or absolute file."""
    relative = abspath("%s/%s" % (prefix_dir, path))
    if os.path.exists(relative):
        return relative

    return abspath(path)


def is_tagged(has_tags, required_tags):
    """Checks if tags match"""
    if len(required_tags) == 0 and len(has_tags) == 0:
        return True
    elif len(required_tags) == 0:
        return False
    else:
        found_tags = []
        for tag in required_tags:
            if tag in has_tags:
                found_tags.append(tag)

        return len(found_tags) == len(required_tags)


def cli_hash(list_of_kv):
    """Parse out a hash from a list of key=value strings"""
    ev_obj = {}
    for ev in list_of_kv:
        ev_list = ev.split('=')
        key = ev_list[0]
        val = '='.join(ev_list[1:])  # b64 and other side effects
        ev_obj[key] = val

    return ev_obj


def merge_dicts(a, b):
    """Deep merge of two dicts"""
    obj = {}
    for key, value in a.iteritems():
        if key in b:
            if isinstance(b[key], dict):
                obj[key] = merge_dicts(value, b.pop(key))
        else:
            obj[key] = value

    for key, value in b.iteritems():
        obj[key] = value

    return obj



""" The CLI interface for aomi """
from __future__ import print_function
import os
import sys

from optparse import OptionParser
import aomi.vault
import aomi.render
import aomi.validation
from aomi.helpers import VERSION as version


def usage():
    """Real Time Help"""
    print("aomi version (%s)" % version)
    print('aomi extract_file <vault path> <file path>')
    print('aomi environment <vault path>')
    print('aomi aws_environment <vault path>')
    print('aomi seed [--secretfile ./Secretfile]'
          ' [--secrets ./secrets] [--policies ./vault]')
    print('aomi template <template> <destination> <path>')
    print('aomi token')


def parser_factory(operation):
    """Return a proper contextual OptionParser"""
    parser = OptionParser()

    parser.add_option('--verbose',
                      dest='verbose',
                      help='Verbose output',
                      action='store_true')
    parser.add_option('--metadata',
                      dest='metadata',
                      help='A series of key=value pairs for token metadata.',
                      default='')
    parser.add_option('--lease',
                      dest='lease',
                      help='Lease time for intermediary token.',
                      default='10s')

    if operation == 'seed':
        parser.add_option('--secrets',
                          dest='secrets',
                          help='Path where secrets are stored',
                          default="%s/.secrets" % os.getcwd())
        parser.add_option('--policies',
                          dest='policies',
                          help='Path where policies are stored',
                          default="%s/vault" % os.getcwd())
        parser.add_option('--secretfile',
                          dest='secretfile',
                          help='Secretfile to use',
                          default="%s/Secretfile" % os.getcwd())
        parser.add_option('--tags',
                          dest='tags',
                          help='Tags of things to seed',
                          default=[],
                          type=str,
                          action='append')
        parser.add_option('--mount-only',
                          dest='mount_only',
                          help='Only mount paths if needed',
                          default=False,
                          action='store_true')
    elif operation == 'environment' or operation == 'template':
        parser.add_option('--add-prefix',
                          dest='add_prefix',
                          help='Specify a prefix to use when '
                          'generating secret key names')
        parser.add_option('--add-suffix',
                          dest='add_suffix',
                          help='Specify a suffix to use when '
                          'generating secret key names')
        parser.add_option('--merge-path',
                          dest='merge_path',
                          action='store_true',
                          default=True,
                          help='merge vault path and key name')
        parser.add_option('--no-merge-path',
                          dest='merge_path',
                          action='store_false',
                          default=True,
                          help='do not merge vault path and key name')
        parser.add_option('--key-map',
                          dest='key_map',
                          action='append',
                          type=str,
                          default=[])

    if operation == 'template' or operation == 'seed':
        parser.add_option('--extra-vars',
                          dest='extra_vars',
                          help='Extra template variables',
                          default=[],
                          type=str,
                          action='append')
        parser.add_option('--extra-vars-file',
                          dest='extra_vars_file',
                          help='YAML files full of variables',
                          default=[],
                          type=str,
                          action='append')

    if operation == 'environment' or operation == 'aws_environment':
        parser.add_option('--export',
                          dest='export',
                          help='Export declared variables',
                          action='store_true')

    if operation == 'environment':
        parser.add_option('--prefix',
                          dest='prefix',
                          help='Old style prefix to use when '
                          'generating secret key names')

    return parser


def parse_extra_vars(extra_vars):
    """Parse out a hash from a list of key=value strings"""
    ev_obj = {}
    for ev in extra_vars:
        key, val = ev.split('=')
        ev_obj[key] = val

    return ev_obj


def action_runner(operation):
    """Run appropriate action, or throw help"""

    parser = parser_factory(operation)
    (opt, args) = parser.parse_args()
    if operation == 'help':
        usage()
        sys.exit(0)

    client = aomi.vault.client(operation, opt)
    if operation == 'extract_file':
        if len(args) == 3:
            aomi.render.raw_file(client, args[1], args[2])
            sys.exit(0)
    elif operation == 'environment':
        if len(args) >= 2:
            paths = args[1:]
            aomi.render.env(client, paths, opt)
            sys.exit(0)
    elif operation == 'aws_environment':
        if len(args) == 2:
            aomi.render.aws(client, args[1], opt)
            sys.exit(0)
    elif operation == 'seed':
        if len(args) == 1:
            aomi.validation.gitignore(opt)
            aomi.vault.seed(client, opt)
            sys.exit(0)
    elif operation == 'template':
        if len(args) >= 4:
            paths = args[3:]
            aomi.render.template(client, args[1], args[2], paths, opt)
            sys.exit(0)
    elif operation == 'token':
        if len(args) == 1:
            print(client.token)
            sys.exit(0)

    usage()
    sys.exit(1)


def main():
    """Entrypoint, sweet Entrypoint"""
    if len(sys.argv) < 2:
        usage()
        sys.exit(1)
    operation = sys.argv[1]
    action_runner(operation)



"""Some validation helpers for aomi"""
from __future__ import print_function
import os
import platform
import stat
from aomi.helpers import problems, abspath, is_tagged, log


def find_file(name, directory):
    """Searches up from a directory looking for a file"""
    path_bits = directory.split(os.sep)
    for i in range(0, len(path_bits) - 1):
        check_path = path_bits[0:len(path_bits) - i]
        check_file = "%s/%s" % (os.sep.join(check_path), name)
        if os.path.exists(check_file):
            return abspath(check_file)

    return None


def subdir_file(item, relative):
    """Returns a file path relative to another file."""
    item_bits = item.split(os.sep)
    relative_bits = relative.split(os.sep)
    for i in range(0, len(item_bits)):
        if i == len(relative_bits) - 1:
            return os.sep.join(item_bits[i:])
        else:
            if item_bits[i] != relative_bits[i]:
                problems("gitignore and secrets paths diverge!")

    return None


def in_file(string, search_file):
    """Looks in a file for a string."""
    h = open(search_file, 'r')
    for l in h.readlines():
        if string in l:
            return True

    return False


def gitignore(opt):
    """Will check directories upwards from the Secretfile in order
    to ensure the gitignore file is set properly"""
    directory = os.path.dirname(abspath(opt.secretfile))
    gitignore_file = find_file('.gitignore', directory)
    if gitignore_file:
        secrets_path = subdir_file(abspath(opt.secrets), gitignore_file)
        if not secrets_path:
            problems("Unable to determine relative location of secretfile")

        if not in_file(secrets_path, gitignore_file):
            problems("The path %s was not found in %s" %
                     (secrets_path, gitignore_file))
    else:
        problems("You should really have a .gitignore")


def secret_file(filename):
    """Will check the permissions of things which really
    should be secret files"""
    filestat = os.stat(abspath(filename))
    if stat.S_ISREG(filestat.st_mode) == 0 and \
       stat.S_ISLNK(filestat.st_mode) == 0:
        problems("Secret file %s must be a real file or symlink" % filename)

    if (platform.system() != "Windows"):
        if filestat.st_mode & stat.S_IROTH or \
           filestat.st_mode & stat.S_IWOTH or \
           filestat.st_mode & stat.S_IWGRP:
            problems("Secret file %s has too loose permissions" % filename)


def validate_obj(keys, obj):
    """Super simple "object" validation."""
    msg = ''
    for k in keys:
        if isinstance(k, str):
            if k not in obj or len(obj[k]) == 0:
                if len(msg) > 0:
                    msg = "%s," % msg

                msg = "%s%s" % (msg, k)
        elif isinstance(k, list):
            found = False
            for k_a in k:
                if k_a in obj:
                    found = True

            if not found:
                if len(msg) > 0:
                    msg = "%s," % msg

                msg = "%s(%s" % (msg, ','.join(k))

    if len(msg) > 0:
        msg = "%s missing" % msg

    return msg


def var_file_obj(obj):
    """Does some validation around a var_file object"""
    check_obj(['var_file', 'mount', 'path'], 'var_file', obj)


def aws_file_obj(obj):
    """Does some validation around an aws_file object"""
    check_obj([['aws_file', 'aws'], 'mount'], 'aws_file', obj)


def aws_secret_obj(filename, obj):
    """Does some validation around AWS secrets"""
    check_obj(['access_key_id', 'secret_access_key'],
              "aws secret %s" % (filename),
              obj)


def file_obj(obj):
    """Basic validation around file objects"""
    check_obj(['mount', 'path'], 'file element', obj)


def tag_check(obj, path, opt):
    """If we require tags, validate for that"""
    if not is_tagged(opt.tags, obj.get('tags', [])):
        log("Skipping %s as it does not have requested tags" %
            path, opt)
        return False
    else:
        return True


def check_obj(keys, name, obj):
    """Do basic validation on an object"""
    msg = validate_obj(keys, obj)
    if len(msg) > 0:
        problems("%s in %s" % (msg, name))


def policy_obj(obj):
    """Basic validation around policy objects"""
    check_obj(['name', 'file'], 'policy', obj)


def user_obj(obj):
    """Do basic validation on a user obj"""
    check_obj(['username', 'password_file', 'policies'],
              'user specification',
              obj)


def audit_log_obj(obj):
    """Do basic validation on an audit log object"""
    check_obj(['type'], 'audit log object', obj)
    if obj['type'] == 'file':
        check_obj(['file_path'], 'file audit log', obj)



""" Render our templates """
import sys
import os
from base64 import b64encode, b64decode
import yaml
from jinja2 import Environment, FileSystemLoader, meta
from aomi.helpers import merge_dicts, problems
import jinja2.nodes


def grok_default_vars(parsed_content):
    """Returns a list of vars for which there is a default being set"""
    default_vars = []
    for node in parsed_content.body[0].nodes:
        if isinstance(node, jinja2.nodes.Filter):
            if node.name == 'default' and node.node.name not in default_vars:
                default_vars.append(node.node.name)

    return default_vars


def render(filename, obj):
    """Render a template, maybe mixing in extra variables"""
    template_path = os.path.abspath(filename)
    fs_loader = FileSystemLoader(os.path.dirname(template_path))
    env = Environment(loader=fs_loader)
    env.filters['b64encode'] = portable_b64encode
    env.filters['b64decode'] = portable_b64decode
    template_src = env.loader.get_source(env, os.path.basename(template_path))
    parsed_content = env.parse(template_src)
    template_vars = meta.find_undeclared_variables(parsed_content)
    if len(template_vars) > 0:
        missing_vars = []
        default_vars = grok_default_vars(parsed_content)
        for var in template_vars:
            if var not in default_vars and var not in obj:
                missing_vars.append(var)

        if len(missing_vars) > 0:
            problems("Missing required variables %s" % ','.join(missing_vars))

    template_obj = env.get_template(os.path.basename(template_path))
    output = template_obj.render(**obj)
    return output


def portable_b64encode(thing):
    """Wrap b64encode for Python 2 & 3"""
    if sys.version_info >= (3, 0):
        return b64encode(bytes(thing, 'utf-8')).decode('utf-8')
    else:
        return b64encode(thing)


def portable_b64decode(thing):
    """Consistent b64decode in Python 2 & 3"""
    if sys.version_info >= (3, 0):
        return b64decode(thing).decode('utf-8')
    else:
        return b64decode(thing)


def load_var_files(opt):
    """Load variable files, merge, return contents"""
    obj = {}
    for var_file in opt.extra_vars_file:
        yamlz = yaml.load(open(os.path.abspath(var_file)).read())
        obj = merge_dicts(obj.copy(), yamlz)

    return obj
